.psize 0, 80
/*
    mbr.s               A Linux native Master Boot Record

    Copyright (C) 2000 Ral Nez de Arenas Coronado
                       <dervishd@linuxfreak.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to FSF:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA 02111-1307 USA

    Report bugs to <dervishd@linuxfreak.com>.
*/

/*
Please note that we are loaded in REAL mode by the BIOS at $0,$0x7c00 and
we need to move out of here, since we need to load the bootsector there.
Moreover, labels won't work as-is, because the assembler, when computing
relative jumps, doesn't know about this displacement.
We move ourselves to $0,$0x0600, which is the very first byte of user
memory available in a PC.

I already know that there are *lots* of labels in the code. I use them
mainly to structure the code, and I like it. They're harmless, anyway ;)
*/

/*
We assemble in the data section since this is not an exe file.
Since we do not provide a section name, it defaults to $0
*/
.data
.code16  /* We assemble in 16 bits real mode (16 bits addressing) */

/* Main program */
.equ START, 0x600               /* Start assembly address */
.equ CTRL,  0x04                /* Bitmask for CTRL  keys in BIOS data area */
.equ ALT,   0x08                /* Bitmask for ALT   keys in BIOS data area */
.equ SHIFT, 0x03                /* Bitmask for SHIFT keys in BIOS data area */
.equ LINUX, 0x83                /* Hex code for Linux partitions */
.equ EMPTY, 0x00                /* Hex code for empty partitions */
.equ BOOTF, 0x80                /* Bitmask for boot flag in partition entry */



.org START  /* Assemble at this address ;) */
    /*
    Be *VERY* careful: the lines of code until the jump to 'start'
    are assembled like if they were on $0,$0x600, but actually they
    are loaded at $0,$0x7c00, so labels WON'T WORK when referencing
    memory addresses (although relative jumps will work).
    */

    jmp entry

/*
Put the strings right here.
This way they are visible when dumping /dev/hd?
*/
idstring: .asciz "DervishD MBR for Linux (LMBR) v@VERSION@\r\n"
message: .asciz "\rChoose partition to boot (1234F): "

entry:
    /* Setup segments */
    cli
    xorw %ax, %ax
    movw %ax, %ss
    xorw %sp, %sp
    movw %ax, %ds
    movw %ax, %es
    sti

    /* Move to $0,$0x0600 */
    cld
    movw $0x7c00, %si           /* Move from here...    */
    movw $0x0600, %di           /* ...to there...       */
    movw $512, %cx              /* ...all our 512 bytes */
    rep movsb                   /* Do the moving        */

    ljmp $0,$start              /* From this point on labels will work */


start:  /* That's the true entry point of the program */
    pushw %dx                   /* Save the boot disk number */

putlogo:  /* Say we're here (words by Neil Turton) ;))) */
    movw $idstring, %si
    call puts

/*
By now, first test is booting the first partition. After that the plan is
that we will run in two modes:
    - normal. Boot the first Linux partition marked as bootable.
    - advanced. The user can choose which primary partition wants to boot
                from. Maybe floppy too?????
*/


/*
If the user pressed any Ctrl key (maybe more will be added, but not by now),
we stopped and present a kind of menu from where the user can choose which
partition to boot or maybe boot from the floppy.
*/
testkey:
    testb $(CTRL), 0x0417  /* Only CTRL by now */
    jnz advancedboot            /* By now it's just normalboot... */

normalboot:
    /* Now we should locate the partition to boot */
    xorw %cx, %cx
nextentry:
    incw %cx
    cmpw $5, %cx
    je advancedboot             /* Revert to advanced mode */

    movw %cx, %bx

    decw %bx                    /* Offset are 0 based */
    shlw $4, %bx                /* Multiply by partition table entry size */

    testb $(BOOTF), table(%bx)  /* Test the boot indicator */
    jz nextentry

    cmpb $(LINUX), table+4(%bx) /* Is it a Linux partition? */
    jne nextentry

    movw %cx, %bx               /* Restore %bx (contains part. number) */
    jmp bootpartition           /* All conditions satisfied: boot partition */


advancedboot:
    xorw %bx, %bx
    movw $message, %si
    call puts
    xor %bx, %bx
getkey:
    /* First, empty keyboard buffer */
    cli
    movw $0x1e, 0x041A          /* Start of keyboard buffer */
    movw $0x1e, 0x041C          /* End of keyboard buffer */
    sti                         /* Keyboard buffer should be clean here */
    xorw %ax, %ax
    int $0x16
    dec %ah
    cmpb $0x01, %ah
    jb getkey
    cmpb $0x04, %ah
    movb %ah, %bl
    jbe bootpartition
    cmpb $0x20, %ah
    je bootfloppy
    jmp getkey


bootpartition:  /* Well, let's boot... */
    popw %dx                    /* Get the boot disk number... */
    pushw %dx                   /* ...and save it again */
    decw %bx
    shlw $4, %bx
    /*
    We assume that %bx points to the correct table entry, not that
    it contains the partition entry number. This translation must be
    done at 'testkey' or 'bootstart'
    */
    inc %bx
    movb table(%bx), %dh
    inc %bx
    movw table(%bx), %cx
    inc %bx
    inc %bx
    movb table(%bx), %al
    cmpb $(EMPTY), %al          /* Skip empty partitions */
    je advancedboot
    jmp readbootsector


bootfloppy:
    xorw %dx, %dx
    movw $0x01, %cx

readbootsector:
    movw $5, %di                /* We will retry 5 times */
retryload:
    decw %di                    /* The Yterbia patch ;) */
    jc advancedboot
    movw $0x7c00, %bx           /* REMEMBER: %ds has the value $0 */
    movw $0x0201, %ax
    int $0x13
    jc retryload                /* Couldn't load bootsector... */

    movw 0x7c00+signature-START, %ax
    cmpw signature, %ax
    jne advancedboot

    /*
    Here we should some cleaning of register contents, but anyway a sane
    bootsector should not rely on the registers having a concrete value,
    except %cs,%ip, of course, which should be $0,$0x7c00 ;)
    */
    ljmp $0, $0x7c00


/* Put a NULL terminated string on the screen */
puts:
    pushw %ax
    pushw %bx
    xorw %bx, %bx
nextchar:
    lodsb
    cmp $0, %al                 /* A '\0' marks the end of the string */
    je return
    movb $0x0e, %ah
    int $0x10
    jmp nextchar
return:
    popw %bx
    popw %ax
    ret

eom:  /* This label marks the end of our MBR */


/* The partition table offset in MBR is $0x01be (446 bytes) */
.org 0x0600+0x01be
table:
.space 0x0040, 0                /* Reserve space for the table */

/* The MBR signature offset is $0x01fe */
.org 0x0600+0x01fe
signature:
.word 0xaa55                    /* MBR signature */
.end  /* Bye */
